#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

#include "configfile.h"
#include "options.h"
#include "term.h"
#include "input.h"
#include "lang.h"

/* command-line options */
#include "hiscore.h"
#include "joystick.h"
#ifdef TWOPLAYER
#include "socket.h"
#endif
#include "game.h"
#include "menu.h"

static void print_term_options()
{
	char colors[8] = "colours";
	char alt[12] = "alternative";
	spellword(colors);
	spellword(alt);
	printf("Terminal options:\n"
"  -width N	set width to N columns (N >= 32)\n"
"  -height N	set height to N lines (N >= 20)\n"
"  -bg black	adjust to dark background\n"
"  -bg white	adjust to light background\n"
"  -color	use %s\n"
"  -mono 	don't use %s\n"
"  -ascii	use ASCII characters for line drawing\n"
"  -vt100	use %s character set for line drawing\n"
"  -block STR	draw blocks using characters in STR\n"
"  -tt		draw blocks as in Mike Taylor's Tetris for Terminals\n"
"  -tt-bg 	tt block characters with background\n",
			colors, colors, alt);
}

static void printhelp()
{
	printf(VITETRIS_VER" ("VERSION_DATE")\n\n"
#ifndef NO_MENU
       "  -nomenu	 skip menus\n"
#endif
       "  -hiscore	 print highscore list and exit\n"
       "  -hiscore FILE  read highscore entries from FILE and merge with\n"
       "		 those in "CONFIG_FILENAME" and "HISCORE_FILENAME"\n"
#ifdef NO_MENU
#  ifdef TWOPLAYER
       "  -inputsetup	 set keys for single player\n"
       "  -inputsetup N  set keys for player N\n"
#  elif defined(JOYSTICK)
       "  -inputsetup	 set keys/buttons\n"
#  else
       "  -inputsetup	 set keys\n"
#  endif
#endif
#ifdef JOYSTICK
       "  -js0 DEVNAME	 joystick device name (e.g. /dev/js0)\n"
       "  -js1 DEVNAME	 second joystick device name\n"
#endif
	"\n"
);
#ifdef SOCKET
	printf("Network play:\n"
#  ifdef INET
       "  -listen PORT		  listen for connections on port PORT\n"
       "  -listen		  listen for local (Unix domain socket) connections\n"
       "  -connect HOSTNAME:PORT  connect to HOSTNAME on PORT\n"
       "			  HOSTNAME may also be an IP address.\n"
       "  -connect PORT		  connect to localhost on PORT\n"
       "  -connect TTY		  connect to TTY (e.g. pts/0)\n"
#  else
       "  -listen	listen for local (Unix domain socket) connections\n"
       "  -connect TTY	connect to TTY (e.g. pts/0)\n"
#  endif
	"\n");
#endif
	printf("Game options:\n"
"  -mode 1	single-player game\n"
"  -mode 1b	single-player game in B-type mode\n"
"  -level 0-9	starting level\n"
"  -height 0-5	height of garbage blocks at beginning\n"
"  -lines N	lines limit (B-type mode)\n"
"  -rotate 0-3	rotation system\n"
"  -softdrop N	softdrop speed (drop N rows)\n"
#ifdef TWOPLAYER
"  -mode 2[b]	two-player game [B-type mode]\n"
"		Player options are prefixed with p1 or p2 (e.g. -p1level 9).\n"
#endif
"\n");
	print_term_options();
	printf("\n"
"Options may also be given as option=ARG, --option=ARG etc.\n");
}

static void set_mode_btype()
{
	union val v;
	v.integ = getopt_int("", "mode") | MODE_BTYPE;
	setoption("", "mode", v, 0);
}

static int proc_game_opt(const char *key, int val, const char *arg)
{
	char sect[8] = "";
	int pl = 0;
	union val v;
#ifdef TWOPLAYER
	if (key[0] == 'p') {
		pl = key[1]-'0';
		if (pl < 1 || pl > 2)
			return 0;
		strcpy(sect, "player");
		sect[6] = key[1];
		key += 2;
	}
#endif
	switch (testgameopt(key, val, pl)) {
	case 0:
		return 0;
	case 1:
		if (arg[1] == 'b')
			val |= MODE_BTYPE;
		break;
	case 4:
		set_mode_btype();
	}
	v.integ = val;
	setoption(sect, key, v, 0);
	return 1;
}

#ifdef JOYSTICK
static void cmd_jsopen(const char *key, char *val)
{
	if (!strcmp(key, "js1"))
		js_open(1, val);
	else
		js_open(0, val);
}
#endif

#ifdef SOCKET
static void cmd_listen(const char *key, char *val)
{
	char *msg;
# ifdef INET
	unsigned port;
	if (val) {
		if (sscanf(val, "%u", &port) < 1)
			return;
		mksocket_inet(NULL, port);
	} else
# endif
	if (msg = mksocket_local(NULL, 1)) {
		puts(msg);
		exit(1);
	}
}

static void cmd_connect(const char *key, char *name)
{
# ifdef INET
	char *p;
	unsigned port;
# endif
	if (!name)
		return;
	while (*name == ' ')
		name++;
# ifndef INET
	connect_tty(name);
# else
	p = strchr(name, ':');
	if (p) {
		*p = '\0';
		p++;
	} else if (!isdigit(*name)) {
		connect_tty(name);
		return;
	} else if (!strchr(name, '.')) {
		p = name;
		name = "localhost";
	}
	if (!p || sscanf(p, "%u", &port) < 1) {
		printf("You need to specify a valid port number "
			"(%s:PORT).\n", name);
		exit(1);
	}
	if (!*name)
		name = "localhost";
	while (1) {
		p = name+strlen(name)-1;
		if (*p != ' ')
			break;
		*p = '\0';
	}
	mksocket_inet(name, port);
# endif
}
#endif /* SOCKET */

static void print_hiscorelist()
{
	char buf[320];
	if (!readhiscores(NULL))
		puts("No saved scores");
	else {
		printf("    Name      Score  Lvl  Lines\n");
		gethiscorelist(buf);
		printf(buf);
	}
}

static void cmd_hiscore(const char *key, char *fname)
{
	if (!fname) {
		print_hiscorelist();
		exit(0);
	}
	readhiscores(fname);
}

static void main_writeconfig()
{
	if (!writeconfig())
		fprintf(stderr, "ERROR! Could not save configuration in "
				CONFIG_FILENAME "\n");
}

#ifdef NO_MENU
static void inputsetup(int pl)
{
	if (pl < 0 || pl > 2)
		pl = 0;
	term_setup();
	atexit(restoreterm);
	init_inputdevs();
	newln(0);
	inputsetup_menu(pl, 2, 1);
	restoreterm();
	main_writeconfig();
	exit(0);
}
#endif

static void setbgdot(union val v, int tp)
{
	if (!tp && v.integ == -1) {
		v.integ = 0;
		setoption("term", "bgdot", v, 1);
	} else if (!getopt_int("term", "bgdot"))
		unsetoption("term", "bgdot");
}

static void proc_term_opt(const char *key, const char *val, int d)
{
	char k[8];
	union val v;
	int tp = 0;
	if (strlen(key) > 7)
		return;
	strcpy(k, key);
	if (val) {
		switch (k[0]) {
		case 'w':
			settermwidth(d);
			return;
		case 'h':
			settermheight(d);
			return;
		case 'b':
			if (!strcmp(k, "bg"))
				v.integ = (d==1 || *val=='w');
			else if (strcmp(k, "block") && strcmp(k, "bgdot"))
				return;
			else {
				strncpy(v.str, val, 4);
				tp = 1;
			}
			goto setopt;
		default:
			v.integ = d;
		}
	} else
		v.integ = 1;
	switch (k[0]) {
	case 'm':
		strcpy(k, "color");
		v.integ = !v.integ;
		break;
	case 'v':
		if (k[1] != 't')
			return;
		v.integ = !v.integ;
	case 'a':
		strcpy(k, "drawing");
		break;
	case 't':
		if (!v.integ) {
			unsetoption("term", "block");
			setbgdot(v, 0);
			return;
		}
		v.integ = -1-(!strcmp(k,"tt-bg") || v.integ==2);
		strcpy(k, "block");
		break;
	default:
		if (strcmp(k, "color"))
			return;
	}
setopt:
	setoption("term", k, v, tp);
	if (!strcmp(k, "block"))
		setbgdot(v, tp);
}

static void proc_cmdline_opt(const char *key, char *val)
{
	const char *ks[] = {
#ifdef JOYSTICK
		"js",
#endif
#ifdef SOCKET
		"listen",
		"connect",
#endif
		"hi"
	};
	void (*handlers[])(const char *key, char *val) = {
#ifdef JOYSTICK
		cmd_jsopen,
#endif
#ifdef SOCKET
		cmd_listen,
		cmd_connect,
#endif
		cmd_hiscore
	};
	int d = 0;
	int i;
	if (val && sscanf(val, "%d", &d) == 1 && proc_game_opt(key, d, val))
		return;
	for (i = 0; 1; i++) {
		if (!strncmp(key, ks[i], strlen(ks[i]))) {
			handlers[i](key, val);
			return;
		}
		if (*ks[i]=='h')
			break;
	}
	if (!strcmp(key, "help") || !val && *key=='h') {
		printhelp();
		exit(0);
	}
#ifdef NO_MENU
	if (!strncmp(key, "inp", 3))
		inputsetup(d);
#else
	if (!strcmp(key, "nomenu"))
		in_menu = val && *val=='0';
#endif
	else
		proc_term_opt(key, val, d);
}

static void proc_args(char **args, int n)
{
	int i, j;
	char *p;
	int c;
	for (i = 0; i < n; i++) {
		j = 0;
		while (args[i][j] == '-')
			j++;
		if (!strlen(args[i]+j))
			continue;
		if (p = strchr(args[i]+j, '=')) {
			*p = '\0';
			p++;
		} else if (i+1 < n) {
			c = args[i+1][0];
			if (c != '-' && (j || !islower(c)
#ifdef SOCKET
					   || !strcmp(args[i], "connect")
#endif
					))
				p = args[i+1];
		}
		proc_cmdline_opt(args[i]+j, p);
		if (p && i+1 < n && p == args[i+1])
			i++;
	}
}

static void cleanup()
{
	freeoptions("");
	input_cleanup();
}

static void finish(int sig)
{
	exit(0);
}

int main(int argc, char **argv)
{
	readoptions();
	atexit(cleanup);
	signal(SIGINT, finish);
	gettermsize();
	getlang();
	if (argc > 1)
		proc_args(argv+1, argc-1);
	init_inputdevs();
	term_startup(socket_fd>0 ? 2 : 0);
#ifndef NO_MENU
	main_writeconfig();
#endif
	return 0;
}
