#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#ifdef CURSES
#undef CURSES
#endif
#include "term.h"
#include "tetris.h"
#include "draw.h"
#include "menu.h"
#include "lang.h"

int term_flags = ASCII;

static char inbuf[4];
static struct termios saved_term_attr;

static char curs_x;
static char curs_y;
static char margin_x;

static char window;
static int win_x;
static int win_y;

#ifdef NO_MENU
#define menuheight 0
#else
static char menuheight;
#endif

void term_setup()
{
	struct termios attr;
	setvbuf(stdin, inbuf, _IOFBF, 4);

	tcgetattr(STDIN_FILENO, &saved_term_attr);
	tcgetattr(STDIN_FILENO, &attr);
	attr.c_lflag &= ~(ICANON|ECHO);
	attr.c_cc[VMIN] = 0;
	attr.c_cc[VTIME] = 2;
	tcsetattr(STDIN_FILENO, TCSADRAIN, &attr);

	printf("\033[?25l");	/* hide cursor */
}

void restoreterm()
{
	if (term_flags == -1)
		return;
	setcurs_end();
	tcsetattr(STDIN_FILENO, TCSANOW, &saved_term_attr);
	printf("\033[?25h\033[m");
	putch('\n');
	term_flags = -1;
}

void setcurs(int x, int y)
{
	x += win_x + margin_x;
	y += win_y;
	if (x != curs_x) {
		if (x > curs_x)
			printf("\033[%dC", x-curs_x);
		else
			printf("\033[%dD", curs_x-x);
		curs_x = x;
	}
	if (y != curs_y) {
		if (y < curs_y)
			printf("\033[%dA", curs_y-y);
		else
			printf("\033[%dB", y-curs_y);
		curs_y = y;
	}
}

void setwcurs(int win, int x, int y)
{
	if (win != window) {
		getwin_xy(win, &win_x, &win_y);
		window = win;
	}
	setcurs(x, y);
}

void movefwd(int n)
{
	printf("\033[%dC", n);
	curs_x += n;
}

void newln(int x)
{
	putchar('\n');
	x += win_x + margin_x;
	if (x)
		printf("\033[%dC", x);
	if (menuheight && curs_x > margin_x+1 && curs_y+1 >= menuheight)
		menuheight = curs_y+1;
	curs_x = x;
	curs_y++;
}

void setcurs_end()
{
	int y;
	if (menuheight)
		y = menuheight;
	else {
		y = term_height-1;
		if (y > 20 && y < 23)
			y = 20;
	}
	setwcurs(0, -margin_x, y);
	fflush(stdout);
}

int is_outside_screen(int x)
{
	return x + win_x + margin_x >= term_width;
}

#ifndef NO_MENU
void get_xy(int *x, int *y)
{
	*x = curs_x - margin_x;
	*y = curs_y;
}
#endif

void refreshwin(int win)
{
	static int redraw;
	struct player *p;
	if (redraw)
		return;
	if ((win == 1 || win == 2) && !game_running) {
		redraw = 1;
		p = &game->player[win-1];
		if (game_paused)
			redrawboard(p, 19);
		else
			redrawboard(p, 3);
		redraw = 0;
	}
	fflush(stdout);
}

/* only used for clearing window showing next tetromino and game screen */
void clearwin(int win)
{
	int h;
	setwcurs(win, 0, 0);
	if (win >= WIN_NEXT)
		clearbox(0, 0, 8, 2);
	else {
		margin_x = 0;
		h = term_height;
		if (h < 24 && h > 21)
			h = 21;
		clearbox(0, 0, 0, h);
	}
}

void cleartoeol()
{
	printf("\033[K");
	if (curs_x <= margin_x+2 && curs_y < menuheight && curs_y > 4)
		menuheight = curs_y;
}

void setcolorpair(int clr)
{
	int bg = -1;
	int bold = 1;
	if (_MONOCHROME) {
		if (clr == MAGENTA_FG)
			setattr_bold();
		return;
	}
	switch (clr) {
	case 7:
		if (_WHITE_BG)
			clr = 0;
		bg = clr;
		break;
	case 0x17:
		printf("\033[m\033[1m");
		return;
	case MAGENTA_FG:
		clr = 5;
		break;
	case WHITE_ON_BLUE:
		clr = 7;
		bg = 4;
		break;
	case BOARD_BG_COLOR:
		clr = (_WHITE_BG) ? 6 : 4;
		bold = 0;
		break;
	case BOARD_FRAME_COLOR:
		clr = 4;
		bold = 0;
		break;
	case PANEL_LABEL_COLOR:
		if (TWOPLAYER_MODE)
			bg = 4;
		else
			bg = (player1.level % 6)+1;
		clr = (bg==6) ? 7 : 3;
		break;
	case RED_FG:
		clr = 1;
		break;
	case YELLOW_ON_BLUE:
		clr = 3;
		bg = 4;
		break;
	case YELLOW_ON_GREEN:
		clr = 3;
		bg = 2;
		break;
	default:
		if (clr & 16)
			clr &= 7;
		else
			bg = clr;
	}
	if (!_TT_BLOCKS) {
		if (bg >= 0)
			printf("\033[4%cm", bg+'0');
		else
			printf("\033[m");
	}
	printf("\033[%c;3%cm", bold+'0', clr+'0');
}

void setattr_normal()
{
	printf("\033[m");
}

void setattr_standout()
{
	printf("\033[7m");
}

void setattr_bold()
{
	if ((term_flags & TT_MONO) != TT_MONO)
		printf("\033[1m");
}

void setattr_underline()
{
	printf("\033[4m");
}

void putch(int ch)
{
	static int acs;
	if (ch < 0)
		ch = (unsigned char) ch;
	if (ch & 0x100) {
		if (_ASCII) {
			putch_ascii(ch);
			return;
		} else if (ch == (' '|0x100) || ch == TEXTURE2 &&
		    (term_flags & (LINUX_TERM | TT_BLOCKS)) != LINUX_TERM)
			ch = ' ';
		else if (!acs) {
			printf("\033(0");
			acs = 1;
		}
		ch &= ~0x100;
	} else if (acs) {
		printf("\033(B");
		acs = 0;
	}
	putchar(ch);
	curs_x += ch=='\b' ? -1 : 1;
}

int printstr(const char *str)
{
	int len = strlen(str);
	int i;
	if (!len)
		return 0;
	if (curs_x+len >= term_width) {
		len = term_width-curs_x-1;
		if (len < 1)
			return 0;
	}
	putch(str[0]);
	for (i = 1; i < len; i++) {
		putchar(str[i]);
		curs_x++;
	}
	return 1;
}

void printint(const char *fmt, int d)
{
	putch(' ');
	putch('\b');
	curs_x += printf(fmt, d);
}

int default_bgdot()
{
	return BULLET;
}

#ifndef NO_MENU
void term_entermenu()
{
	while (curs_y < 4)
		newln(0);
	menuheight = 4;
	margin_x = getmargin_x();
	draw_tetris_logo(0, 0);
}

static int op_handler(int k, int *pos)
{
	const struct termopt opts[3] = {
		{ WHITE_BG, "black white", "bg" },
		{ MONOCHROME, "16 mono", "color" },
		{ ASCII, "VT100 ASCII", "drawing" }
	};
	int i = *pos-2;
	if (i < 0) {
		if (k)
			inputsetup_box(k-1, 1, 4);
		return 1;
	}
	if (i < 2+!(term_flags & CYGWIN))
		i = term_optionhandler(k, opts+i);
	else {
		if (!k)
			menuheight = 11;
		i = term_blockstyle(k);
	}
	if (i == 3)
		draw_tetris_logo(0, 0);
	return i;
}

static void term_optionsmenu()
{
	const char *items[3] = {"Term BG", NULL, "Line Drawing"};
	char colors[8] = "Colours";
	spellword(colors);
	items[1] = colors;
	optionsmenu(items, 2+!(term_flags & CYGWIN), op_handler, 1, 4);
	menuheight = 8;
}

static void hiscorelist()
{
	menuheight = 6;
	show_hiscorelist(1, 4);
	clearbox(0, 4, 0, menuheight-4);
}

#ifdef TWOPLAYER
int term_startupmenu(int i)
{
	while (1) {
		print_vitetris_ver(19, 4);
		newln(-margin_x);
		newln(19);
		printstr("Written by"); newln(19);
		printstr("Victor Nilsson"); newln(19);
		printstr("2007-2008");
		i = startupmenu(i-1, 1, 4);
# ifndef SOCKET
		if (!i)
			break;
# else
		if (i==2 && !select_2p_tty(17, 5))
			continue;
		if (!i && (i = menu_checkinvit(1, 4)) == -1) {
			i = 1;
			continue;
		}
		if (!i)
			break;
		setcurs_end();
		printf("\033[K");
# endif
		clearbox(0, 4, 0, 6);
		if (i==5)
			term_optionsmenu();
		else if (i==6)
			hiscorelist();
		else
			break;
	}
	return i;
}
#endif /* TWOPLAYER */

int term_gamemenu()
{
	int i = 1;
#ifdef TWOPLAYER
	int h;
	if (!TWOPLAYER_MODE || game->mode & MODE_NETWORK)
		h = GAMEMENU_LENGTH+4;
	else {
		h = 17;
		while (menuheight < h)
			newln(2);
	}
	while (1) {
		i = gamemenu(i-1, 1, 4);
		if (i==4)
			inputsetup_box(0, 1, 4);
		else
			break;
	}
	if (!i)
		clearbox(0, 4, 0, h-4);
#else
	while (1) {
		i = gamemenu(i-1, 1, 4);
		if (i==4) {
			setcurs(0, 13);
			printf("\033[K");
			inputsetup_box(0, 1, 4);
		} else if (i < GAMEMENU_LENGTH-1)
			break;
		else {
			clearbox(0, 4, 0, GAMEMENU_LENGTH);
			if (i == GAMEMENU_LENGTH-1)
				term_optionsmenu();
			else
				hiscorelist();
		}
	}
#endif
	return i;
}
#endif /* !NO_MENU */

void term_entergame()
{
	menuheight = 0;
	clearwin(0);
	margin_x = getmargin_x();
}
